/* client.c */

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <errno.h>

#if defined(__linux__)
#include <unistd.h>
#include <sys/poll.h>
#endif

#if defined(__CYGWIN__)
#include <sys/poll.h>
#endif

#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glut.h>
#include <math.h>

#include "conn.h"
#include "game.h"

static conn_t con;
static state_t state;
static coord_t pos;
static player_t player;
static struct pollfd fds[1];
static int timeout = 20;

static void move(const coord_t * from, const coord_t * to)
{
	move_t move;

	assert(to);
	assert(con.sock >= 0);

	move.from = *from;
	move.to = *to;

	state_move_player(&state, from, to, player);

	if (conn_write(&con, &move, sizeof(move)) != sizeof(move)) {
		printf("error: cannot send move.\n");
		perror("write");
		exit(-1);
	}
}

static void display(void)
{
	int x, y;

	glClear(GL_COLOR_BUFFER_BIT);
	glLoadIdentity();

	/* draw grid */
	glColor4f(1.0, 1.0, 1.0, 1.0);
	glBegin(GL_LINES);
	for (y = 0; y <= FIELD_N; ++y) {
		glVertex2i(0, y * 20);
		glVertex2i(20 * FIELD_N, y * 20);
	}
	for (x = 0; x <= FIELD_N; ++x) {
		glVertex2i(x * 20, 0);
		glVertex2i(x * 20, 20 * FIELD_N);
	}
	glEnd();

	/* draw players */
	for (y = 0; y < FIELD_N; ++y) {
		for (x = 0; x < FIELD_N; ++x) {
			player_t player = state_get(&state, x, y);
			if (player == 0) continue;
			switch (player) {
				case 1: glColor4f(1.0, 0.0, 0.0, 1.0); break;
				case 2: glColor4f(0.0, 1.0, 0.0, 1.0); break;
				case 3: glColor4f(0.0, 0.0, 1.0, 1.0); break;
				case 4: glColor4f(1.0, 1.0, 0.0, 1.0); break;
			}
			glBegin(GL_QUADS);
				glVertex2i((x+0) * 20 + 1, (y+0) * 20 + 0);
				glVertex2i((x+0) * 20 + 1, (y+1) * 20 - 1);
				glVertex2i((x+1) * 20 - 0, (y+1) * 20 - 1);
				glVertex2i((x+1) * 20 - 0, (y+0) * 20 + 0);
			glEnd();
		}
	}

	glFlush();
	glutSwapBuffers();
}

static void reshape(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0, w, h, 0, 0, 1);
	glMatrixMode(GL_MODELVIEW);
}

static void idle(void)
{
	/* poll connection */
	if (poll(fds, 1, timeout)) {
		if (fds[0].revents & POLLIN) {
			if (conn_read(&con, &state, sizeof(state)) != sizeof(state)) {
				perror("read");
				exit(-1);
			}
		}
	}
	glutPostRedisplay();
}

static void keyboard(unsigned char key, int x, int y)
{
	coord_t old_pos = pos;
	coord_t new_pos = pos;
	int mov = 0;

	switch (key) {
		case 27: exit(0);
		case 'w':
			if (pos.y > 0) { --new_pos.y; mov = 1; }
			break;
		case 's':
			if (pos.y < FIELD_N-1) { ++new_pos.y; mov = 1; }
			break;
		case 'a':
			if (pos.x > 0) { --new_pos.x; mov = 1; }
			break;
		case 'd':
			if (pos.x < FIELD_N-1) { ++new_pos.x; mov = 1; }
			break;
	}
	if (mov) {
		if (state_get_coord(&state, &new_pos) == 0) {
			pos = new_pos;
			move(&old_pos, &new_pos);
			glutPostRedisplay();
		}
	}
}

static void cleanup(void)
{
	conn_close(&con);
}

int main(int argc, char ** argv)
{
	atexit(cleanup);
	state_init(&state);

	/* establish connection to server */
	if (conn_open(&con, 0x7f000001, 9900) < 0) {
		printf("error: cannot connect to server.\n");
		perror("open");
		exit(-1);
	}
	if (conn_read(&con, &player, sizeof(player)) != sizeof(player)) {
		printf("error: cannot receive player number.\n");
		perror("read");
		exit(-1);
	}
	if (conn_read(&con, &pos, sizeof(pos)) != sizeof(pos)) {
		printf("error: cannot receive position.\n");
		perror("read");
		exit(-1);
	}
	if (conn_read(&con, &state, sizeof(state)) != sizeof(state)) {
		printf("error: cannot receive game state.\n");
		perror("read");
		exit(-1);
	}

	/* information about client */
	printf("player: %d\n", player);
	printf("pos   : { %d , %d }\n", pos.x, pos.y);

	/* prepare structure for poll */
	fds[0].fd = con.sock;
	fds[0].events = POLLIN;

	/* init GL */
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_ALPHA);
	glutInitWindowPosition(0, 0);
	glutInitWindowSize(320, 320);
	glutCreateWindow("Multiplayer");
	glutDisplayFunc(display);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyboard);
	glutIdleFunc(idle);
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glFrontFace(GL_CCW);
	glEnable(GL_NORMALIZE);
	glPolygonMode(GL_FRONT, GL_FILL);
	glPolygonMode(GL_BACK, GL_LINE);
	glutMainLoop();
	return 0;
}

