#include "conn.h"
#include <stdio.h>
#include <assert.h>
#include <errno.h>
#include <string.h>

#if defined(__linux__)
#include <sys/socket.h>
#include <unistd.h>
#endif

static int wait_to_read(int sock, struct timeval * t) /* {{{ */
{
	int rc;
	fd_set fds;

	assert(sock >= 0);
	assert(t);
	FD_ZERO(&fds);
	FD_SET(sock, &fds);
	rc = select(sock+1, &fds, NULL, NULL, t);
	if (rc == 0) return -2;
	if (rc < 0) return -1;
	return 0;
} /* }}} */

void server_clear(server_t * srv) /* {{{ */
{
	assert(srv);
	memset(srv, 0, sizeof(server_t));
	srv->sock = -1;
} /* }}} */

void server_clear_all(server_t * srv, const int n) /* {{{ */
{
	int i;

	assert(srv);
	assert(n >= 0);
	for (i = 0; i < n; ++i) server_clear(srv+i);
} /* }}} */

int server_open(server_t * s, long ip, short port) /* {{{ */
{
	int t = 1;
	int sock = -1;

	assert(s);

	s->sock = -1;
	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("socket");
		goto error;
	}

	if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &t, sizeof(t)) < 0) {
		perror("setsockopt");
		goto error;
	}

	memset(&s->addr, 0, sizeof(s->addr));
	s->addr.sin_family = AF_INET;
	s->addr.sin_port = htons(port);
	s->addr.sin_addr.s_addr = htonl(ip);
	if (bind(sock, (const struct sockaddr *)&s->addr, sizeof(s->addr)) < 0) {
		perror("bind");
		goto error;
	}

	if (listen(sock, 1) < 0){
		perror("listen");
		goto error;
	}

	s->sock = sock;
	return 0;
error:
	if (sock >= 0) close(sock);
	return -1;
} /* }}} */

void server_close(server_t * s) /* {{{ */
{
	assert(s);
	if (s->sock >= 0) {
		close(s->sock);
		s->sock = -1;
	}
} /* }}} */

void server_close_all(server_t * srv, const int n) /* {{{ */
{
	int i;

	assert(srv);
	assert(n >= 0);
	for (i = 0; i < n; ++i) server_close(srv+i);
} /* }}} */

void conn_clear(conn_t * c) /* {{{ */
{
	assert(c);
	memset(c, 0, sizeof(conn_t));
	c->sock = -1;
} /* }}} */

void conn_close(conn_t * c) /* {{{ */
{
	assert(c);
	if (c->sock >= 0) {
		close(c->sock);
		conn_clear(c);
	}
} /* }}} */

int conn_open(conn_t * c, long ip, short port) /* {{{ */
{
	int sock = -1;
	socklen_t len;

	assert(c);

	c->sock = -1;

	sock = socket(AF_INET, SOCK_STREAM, 0);
	if (sock < 0) {
		perror("socket");
		goto error;
	}

	memset(&c->addr, 0, sizeof(c->addr));
	c->addr.sin_family = AF_INET;
	c->addr.sin_port = htons(port);
	c->addr.sin_addr.s_addr = htonl(ip);
	if (connect(sock, (const struct sockaddr *)&c->addr, sizeof(c->addr)) < 0) {
		perror("connect");
		goto error;
	}

	len = sizeof(c->local);
	if (getsockname(sock, (struct sockaddr *)&c->local, &len) < 0) {
		perror("getsockname");
		goto error;
	}

	c->sock = sock;
	return 0;
error:
	if (sock >= 0) close(sock);
	return -1;
} /* }}} */

int conn_accept(server_t * s, conn_t * c) /* {{{ */
{
	socklen_t len;
	int sock = -1;

	assert(s);
	assert(c);
	len = sizeof(c->addr);
	sock = accept(s->sock, (struct sockaddr *)&c->addr, &len);
	if (sock < 0) {
		perror("accept");
		goto error;
	}

	len = sizeof(c->local);
	if (getsockname(sock, (struct sockaddr *)&c->local, &len) < 0) {
		perror("getsockname");
		goto error;
	}

	c->sock = sock;
	return 0;
error:
	if (sock >= 0) close(sock);
	return -1;
} /* }}} */

void conn_clear_all(conn_t * c, const int n) /* {{{ */
{
	int i;

	assert(c);
	assert(n >= 0);
	for (i = 0; i < n; ++i) conn_clear(c+i);
} /* }}} */

void conn_close_all(conn_t * a, const int n) /* {{{ */
{
	int i;

	assert(a);
	assert(n >= 0);
	for (i = 0; i < n; ++i) conn_close(a+i);
} /* }}} */

int conn_add(conn_t * a, const int n, conn_t * c) /* {{{ */
{
	int i;

	assert(a);
	assert(n >= 0);
	assert(c);
	for (i = 0; i < n; ++i) {
		if (a[i].sock < 0) {
			a[i] = *c;
			return 0;
		}
	}
	return -1;
} /* }}} */

int conn_close_del(conn_t * a, const int n, conn_t * c) /* {{{ */
{
	int i;

	assert(a);
	assert(n >= 0);
	assert(c);
	for (i = 0; i < n; ++i) {
		if (a[i].sock == c->sock) {
			conn_close(a+i);
			return 0;
		}
	}
	return -1;
} /* }}} */

int conn_wait_to_read(conn_t * c, struct timeval * t) /* {{{ */
{
	assert(c);
	assert(t);
	return wait_to_read(c->sock, t);
} /* }}} */

int conn_write(conn_t * c, const void * buf, int len) /* {{{ */
{
	assert(c);
	assert(c->sock >= 0);
	assert(buf);
	assert(len > 0);
	return write(c->sock, buf, len);
} /* }}} */

int conn_read(conn_t * c, void * buf, int len) /* {{{ */
{
	int rc;

	assert(c);
	assert(c->sock >= 0);
	assert(buf);
	assert(len > 0);
	return read(c->sock, buf, len);
} /* }}} */

