/* srv.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#if defined(__linux__)
#include <sys/select.h>
#include <unistd.h>
#endif

#include "conn.h"

static void prepare_fds(fd_set * fds,
	conn_t * arr_conn, const int n_conn,
	server_t * arr_srv, const int n_srv,
	int * fd_max)
{
	int i;
	int m = -1;

	assert(fds);
	assert(arr_conn);
	assert(n_conn >= 0);
	assert(arr_srv);
	assert(n_srv >= 0);
	assert(fd_max);

	FD_ZERO(fds);
	for (i = 0; i < n_srv; ++i) { /* add all servers */
		server_t * s = arr_srv+i;
		if (s->sock >= 0) {
			if (m < s->sock) m = s->sock;
			FD_SET(s->sock, fds);
		}
	}
	for (i = 0; i < n_conn; ++i) { /* add all connections */
		conn_t * c = arr_conn+i;
		if (c->sock >= 0) {
			if (m < c->sock) m = c->sock;
			FD_SET(c->sock, fds);
		}
	}
	*fd_max = m + 1;
}

enum { MAX_CONN = 64 };

static server_t srv;
static conn_t con[MAX_CONN];
static int n_cnt = 0;
static fd_set fds;

static void close_connection(conn_t * c)
{
	if (conn_close_del(con, MAX_CONN, c) < 0) {
		printf("cannot remove connection from list\n");
	} else {
		--n_cnt;
	}
}

static int new_connection(server_t * s)
{
	assert(s);
	conn_t c;

	if (conn_accept(s, &c) < 0) return -1;
	if (n_cnt < MAX_CONN) {
		if (conn_add(con, MAX_CONN, &c) < 0) {
			conn_close(&c);
			printf("not able to handle new connection\n");
			return -1;
		} else {
			++n_cnt;
		}
	} else {
		/* this is necessary to let the other know
		   the connection is closed. */
		conn_close(&c);
		printf("cannot accept more connections\n");
		return -1;
	}
	return 0;
}

static void handle_connection(conn_t * c)
{
	int rc;
	char buf[128];

	assert(c);
	assert(c->sock >= 0);

	memset(buf, 0, sizeof(buf));
	rc = conn_read(c, buf, sizeof(buf));
	if (rc == 0) goto close_conn;
	if (rc < 0) {
		if (errno == ECONNABORTED) goto close_conn;
		perror("read");
		goto close_conn;
	}
	rc = conn_write(c, buf, sizeof(buf));
	if (rc != sizeof(buf)) {
		printf("error while writing to connection.\n");
		goto close_conn;
	}

	return;
close_conn:
	close_connection(c);
	return;
}

static void cleanup(void)
{
	conn_close_all(con, MAX_CONN);
	server_close(&srv);
}

int main(int argc, char ** argv)
{
	int fd_max = -1;
	int rc;
	int i;
	short port;

	atexit(cleanup);
	server_clear(&srv);
	conn_clear_all(con, MAX_CONN);

	port = (argc < 2) ? 9900 : atoi(argv[1]);
	if (server_open(&srv, INADDR_ANY, port)) {
		printf("error: cannot open server.\n");
		return -1;
	}
	printf("server ready.\n");

	for (;;) {
		prepare_fds(&fds, con, MAX_CONN, &srv, 1, &fd_max);
		rc = select(fd_max, &fds, NULL, NULL, NULL);
		if (rc < 0) {
			perror("select");
			printf("errno=%d\n", errno);
			server_close(&srv);
			conn_close_all(con, MAX_CONN);
			return -1;
		}
		/* check server */
		if ((srv.sock >= 0) && FD_ISSET(srv.sock, &fds)) {
			if (new_connection(&srv) < 0) {
				printf("error: could not open/add new connection.\n");
			}
			--rc;
		}
		/* check all connections */
		for (i = 0; rc && (i < MAX_CONN); ++i) {
			conn_t * c = con+i;
			if ((c->sock >= 0) && FD_ISSET(c->sock, &fds)) {
				handle_connection(c);
				--rc;
			}
		}
	}
	return 0;
}

