/* srv.c */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#if defined(__linux__)
#include <sys/select.h>
#include <unistd.h>
#endif

#include "conn.h"
#include "game.h"

static void prepare_fds(fd_set * fds,
	conn_t * arr_conn, const int n_conn,
	server_t * arr_srv, const int n_srv,
	int * fd_max)
{
	int i;
	int m = -1;

	assert(fds);
	assert(arr_conn);
	assert(n_conn >= 0);
	assert(arr_srv);
	assert(n_srv >= 0);
	assert(fd_max);

	FD_ZERO(fds);
	for (i = 0; i < n_srv; ++i) { /* add all servers */
		server_t * s = arr_srv+i;
		if (s->sock >= 0) {
			if (m < s->sock) m = s->sock;
			FD_SET(s->sock, fds);
		}
	}
	for (i = 0; i < n_conn; ++i) { /* add all connections */
		conn_t * c = arr_conn+i;
		if (c->sock >= 0) {
			if (m < c->sock) m = c->sock;
			FD_SET(c->sock, fds);
		}
	}
	*fd_max = m + 1;
}

enum { MAX_CONN = MAX_PLAYERS };

static server_t srv;
static conn_t con[MAX_CONN];
static int n_cnt = 0;
static fd_set fds;

static state_t state;

static void close_connection(conn_t * c)
{
	if (conn_close_del(con, MAX_CONN, c) < 0) {
		printf("cannot remove connection from list\n");
	} else {
		--n_cnt;
	}
}

static int send_state(conn_t * c)
{
	assert(c);

	if (conn_write(c, &state, sizeof(state)) != sizeof(state))
		return -1;
	return 0;
}

static void send_state_to_all_except(conn_t * c)
{
	int i;

	for (i = 0; i < MAX_CONN; ++i) {
		conn_t * co = con+i;
		if (co->sock == -1) continue;
		if (c && co->sock == c->sock) continue;
		if (send_state(co)) {
			printf("error while writing to connection.\n");
			perror("write");
			close_connection(co);
		}
	}
}

static int send_game_info(conn_t * c)
{
	player_t player;
	coord_t pos;
	int i;

	assert(c);

	/* determine player ID */
	player = -1;
	for (i = 0; i < MAX_CONN; ++i) {
		if (con[i].sock == c->sock) {
			player = i;
			break;
		}
	}
	if (player < 0) return -1;

	/* send player ID to client */
	++player;
	if (conn_write(c, &player, sizeof(player)) != sizeof(player))
		return -1;

	/* determine random coordinates */
	coord_init(&pos);
	for (;;) {
		pos.x = rand() % FIELD_N;
		pos.y = rand() % FIELD_N;
		if (state_get_coord(&state, &pos) == 0) {
			state_set_coord(&state, &pos, player);
			break;
		}
	}

	/* send coordingates to client */
	if (conn_write(c, &pos, sizeof(pos)) != sizeof(pos)) goto error;

	/* send game state to client */
	if (send_state(c)) goto error;
	return 0;
error:
	state_set_coord(&state, &pos, 0);
	return -1;
}

static int new_connection(server_t * s)
{
	assert(s);
	conn_t c;

	if (conn_accept(s, &c) < 0) return -1;
	if (n_cnt < MAX_CONN) {
		if (conn_add(con, MAX_CONN, &c) < 0) {
			conn_close(&c);
			printf("not able to handle new connection\n");
			return -1;
		} else {
			++n_cnt;
			if (send_game_info(&c)) {
				--n_cnt;
				conn_close(&c);
				return -1;
			} else {
				send_state_to_all_except(&c);
			}
		}
	} else {
		/* this is necessary to let the other know
		   the connection is closed. */
		conn_close(&c);
		printf("cannot accept more connections\n");
		return -1;
	}
	return 0;
}

static void handle_connection(conn_t * c)
{
	int rc;
	int i;
	move_t move;
	player_t player;

	assert(c);
	assert(c->sock >= 0);

	memset(&move, 0, sizeof(move));
	rc = conn_read(c, &move, sizeof(move));
	if (rc == 0) goto close_conn;
	if (rc < 0) {
		if (errno == ECONNABORTED) goto close_conn;
		perror("read");
		goto close_conn;
	}

	/* perform move */
	player = state_get_coord(&state, &move.from);
	if (player) {
		state_set_coord(&state, &move.to, player);
		state_set_coord(&state, &move.from, 0);
		send_state_to_all_except(c);
	}

	return;
close_conn:
	close_connection(c);
	player = -1;
	for (i = 0; i < MAX_CONN; ++i) {
		if (con[i].sock == c->sock) {
			player = i;
			break;
		}
	}
	if (player >= 0) state_remove_player(&state, player+1);
	send_state_to_all_except(NULL);
	return;
}

static void cleanup(void)
{
	conn_close_all(con, MAX_CONN);
	server_close(&srv);
}

int main(int argc, char ** argv)
{
	int fd_max = -1;
	int rc;
	int i;

	srand(time(0));
	atexit(cleanup);
	server_clear(&srv);
	conn_clear_all(con, MAX_CONN);
	state_init(&state);

	if (server_open(&srv, INADDR_ANY, 9900)) {
		printf("error: cannot open server.\n");
		return -1;
	}
	printf("server ready.\n");

	for (;;) {
		prepare_fds(&fds, con, MAX_CONN, &srv, 1, &fd_max);
		rc = select(fd_max, &fds, NULL, NULL, NULL);
		if (rc < 0) {
			perror("select");
			printf("errno=%d\n", errno);
			server_close(&srv);
			conn_close_all(con, MAX_CONN);
			return -1;
		}
		/* check server */
		if ((srv.sock >= 0) && FD_ISSET(srv.sock, &fds)) {
			if (new_connection(&srv) < 0) {
				printf("error: could not open/add new connection.\n");
			}
			--rc;
		}
		/* check all connections */
		for (i = 0; rc && (i < MAX_CONN); ++i) {
			conn_t * c = con+i;
			if ((c->sock >= 0) && FD_ISSET(c->sock, &fds)) {
				handle_connection(c);
				--rc;
			}
		}
	}
	return 0;
}

